/**
 * @license
 * Copyright Google LLC All Rights Reserved.
 *
 * Use of this source code is governed by an MIT-style license that can be
 * found in the LICENSE file at https://angular.io/license
 */

import {assertInInjectionContext} from '../../di/contextual';
import {Injector} from '../../di/injector';
import {inject} from '../../di/injector_compatibility';
import {DestroyRef} from '../../linker/destroy_ref';
import {Watch} from '../../signals';

/**
 * An effect can, optionally, return a cleanup function. If returned, the cleanup is executed before
 * the next effect run. The cleanup function makes it possible to "cancel" any work that the
 * previous effect run might have started.
 *
 * @developerPreview
 */
export type EffectCleanupFn = () => void;

const globalWatches = new Set<Watch>();
const queuedWatches = new Map<Watch, Zone>();

let watchQueuePromise: {promise: Promise<void>; resolveFn: () => void;}|null = null;

/**
 * A global reactive effect, which can be manually destroyed.
 *
 * @developerPreview
 */
export interface EffectRef {
  /**
   * Shut down the effect, removing it from any upcoming scheduled executions.
   */
  destroy(): void;
}

/**
 * Options passed to the `effect` function.
 *
 * @developerPreview
 */
export interface CreateEffectOptions {
  /**
   * The `Injector` in which to create the effect.
   *
   * If this is not provided, the current injection context will be used instead (via `inject`).
   */
  injector?: Injector;

  /**
   * Whether the `effect` should require manual cleanup.
   *
   * If this is `false` (the default) the effect will automatically register itself to be cleaned up
   * with the current `DestroyRef`.
   */
  manualCleanup?: boolean;
}

/**
 * Create a global `Effect` for the given reactive function.
 *
 * @developerPreview
 */
export function effect(
    effectFn: () => EffectCleanupFn | void, options?: CreateEffectOptions): EffectRef {
  !options?.injector && assertInInjectionContext(effect);

  const zone = Zone.current;
  const watch = new Watch(effectFn, (watch) => queueWatch(watch, zone));

  const injector = options?.injector ?? inject(Injector);
  const destroyRef = options?.manualCleanup !== true ? injector.get(DestroyRef) : null;

  globalWatches.add(watch);

  // Effects start dirty.
  watch.notify();

  let unregisterOnDestroy: (() => void)|undefined;

  const destroy = () => {
    watch.cleanup();
    unregisterOnDestroy?.();
    queuedWatches.delete(watch);
    globalWatches.delete(watch);
  };

  unregisterOnDestroy = destroyRef?.onDestroy(destroy);

  return {
    destroy,
  };
}

function queueWatch(watch: Watch, zone: Zone): void {
  if (queuedWatches.has(watch) || !globalWatches.has(watch)) {
    return;
  }

  queuedWatches.set(watch, zone);

  if (watchQueuePromise === null) {
    Promise.resolve().then(runWatchQueue);

    let resolveFn!: () => void;
    const promise = new Promise<void>((resolve) => {
      resolveFn = resolve;
    });

    watchQueuePromise = {
      promise,
      resolveFn,
    };
  }
}

function runWatchQueue(): void {
  for (const [watch, zone] of queuedWatches) {
    queuedWatches.delete(watch);
    zone.run(() => watch.run());
  }

  watchQueuePromise!.resolveFn();
  watchQueuePromise = null;
}
